import pytest
from carbon_index.node import TrieNode
from carbon_index.index import CarbonIndex


@pytest.fixture
def trie_node():
    # level 1
    ZG = TrieNode('ZG', is_leaf=False)
    # level 2
    seattle = TrieNode('Seattle', is_leaf=False)
    new_york = TrieNode('NewYork', is_leaf=False)
    san_francisco = TrieNode('SF', is_leaf=False)
    # level 3
    zillow = TrieNode('zillow')
    zillow2 = TrieNode('zillow2')
    zillow3 = TrieNode('zil3low')
    hotpads = TrieNode('hotpads')
    trulia = TrieNode('trulia')
    streeteasy = TrieNode('streeteasy')
    # relations
    seattle.add(zillow)
    seattle.add(zillow2)
    seattle.add(zillow3)
    new_york.add(hotpads)
    new_york.add(streeteasy)
    san_francisco.add(trulia)
    ZG.add(seattle)
    ZG.add(new_york)
    ZG.add(san_francisco)
    return ZG


@pytest.fixture
def carbon_index():
    index = CarbonIndex()
    index.insert('ZG.zillow.velocity')
    index.insert('ZG.zillow.velocity.mondev')
    index.insert('ZG.zillow.product.rental')
    index.insert('ZG.zillow.product.listing')
    index.insert('ZG.trulia.product')
    index.insert('ZG.trulia.product.rental')
    return index



from .node import TrieNode


class CarbonIndex:
    """
    an index for carbon-cache instances.
    """

    def __init__(self, name='carbon_index', sep='.'):
        self.name = name
        self.root = TrieNode('root', is_leaf=False)
        self.sep = sep

    def insert(self, metric):
        """
        insert metric to carbon Trie Index.

        args: metric name.
        """
        metric_parts = metric.split(self.sep)
        self._insert(self.root, metric_parts)

    def has_metric(self, metric):
        """
        check if the metric exists.

        args: metric.
        """
        metric_parts = metric.split(self.sep)
        cur = self.root
        for part in metric_parts:
            if not cur.has_child(part):
                return False
            cur = cur.get(part)
        return cur.is_leaf

    def delete(self, metric):
        """
        remove a metric from index. wild.

        args: metric, should not include wildcards.

        return: True/False.
        """
        metric_parts = metric.split(self.sep)
        return self._delete(self.root, metric_parts)

    def expand_query(self, query):
        """
        expand a wildcard query
        """
        return self.root.expand_query(query)

    def _insert(self, parent, metric_parts):
        """
        a private helper function for insert metric.
        """
        if len(metric_parts) == 0:
            return
        if len(metric_parts) == 1:
            parent.add(TrieNode(metric_parts[0]))
            return
        if not parent.get(metric_parts[0]):
            parent.add(TrieNode(metric_parts[0], is_leaf=False))
        self._insert(parent.get(metric_parts[0]), metric_parts[1:])

    def _delete(self, cur, metric_parts):
        if len(metric_parts) == 0:
            return cur.is_leaf
        if cur.has_child(metric_parts[0]):
            nxt = cur.get(metric_parts[0])
            deleted = self._delete(nxt, metric_parts[1:])
            if nxt.count() == 0:
                cur.delete(nxt.name)
            return deleted
        else:
            return False



from carbon_index.node import TrieNode


def test_node_get(trie_node):
    seattle = trie_node.get('Seattle')
    assert seattle is not None
    assert seattle.get('zillow') is not None
    assert seattle.get('zillow2') is not None
    assert seattle.get('zil3low') is not None


def test_node_has_child(trie_node):
    assert trie_node.has_child('Seattle') is True
    assert trie_node.has_child('NewYork') is True
    assert trie_node.has_child('SF') is True
    assert trie_node.has_child('Boston') is False


def test_node_get_all_wildcard_star(trie_node):
    all_matches = trie_node.get_all('*')
    all_match_strs = [match.name for match in all_matches]
    assert sorted(all_match_strs) == ['NewYork', 'SF', 'Seattle']


def test_node_get_all_wildcard_star_partial(trie_node):
    seattle = trie_node.get('Seattle')
    all_matches = seattle.get_all('zill*')
    all_match_strs = [match.name for match in all_matches]
    assert sorted(all_match_strs) == ['zillow', 'zillow2']


def test_node_get_all_wildcard_brackets(trie_node):
    seattle = trie_node.get('Seattle')
    all_matches = seattle.get_all('zil[0-9]low')
    assert len(all_matches) == 1
    assert 'zil3low' == all_matches[0].name


def test_node_get_all_wildcard_braces(trie_node):
    seattle = trie_node.get('Seattle')
    all_matches = seattle.get_all('zill{ow,ow2}')
    all_match_strs = [match.name for match in all_matches]
    assert sorted(all_match_strs) == ['zillow', 'zillow2']


def test_node_add(trie_node):
    assert trie_node.get('LA') is None
    trie_node.add(TrieNode('LA'))
    assert trie_node.has_child('LA')


def test_node_add_is_leaf(trie_node):
    assert trie_node.get('SF').is_leaf is False
    trie_node.add(TrieNode('SF'))
    assert trie_node.get('SF').is_leaf is True


def test_expand_query_no_wildcard(trie_node):
    queries = trie_node.expand_query('Seattle.zillow')
    assert queries == ['Seattle.zillow']


def test_expand_query_not_exist(trie_node):
    queries = trie_node.expand_query('Seattle.trulia')
    assert queries == []


def test_expand_query_wildcards(trie_node):
    queries = trie_node.expand_query('Seattle.*')
    assert len(queries) == 3
    queries = trie_node.expand_query('*.*')
    assert len(queries) == 6
    assert 'NewYork.hotpads' in queries


def test_is_exist(trie_node):
    assert trie_node.is_exist('Seattle.zillow') is True
    assert trie_node.is_exist('NewYork.hotpads') is True
    assert trie_node.is_exist('Seattle.trulia') is False


def test_is_exist_add_leaf_in_branch_node(trie_node):
    assert trie_node.is_exist('Seattle.zillow') is True
    assert trie_node.is_exist('Seattle') is False
    trie_node.add(TrieNode('Seattle'))
    assert trie_node.is_exist('Seattle') is True



import fnmatch
from .utils import expand_braces


class TrieNode:
    """
    an standard unit in carbon index.
    """

    def __init__(self, name, is_leaf=True, sep="."):
        self.is_leaf = is_leaf
        self.name = name
        self.sep = "."
        self.children = dict()

    def get(self, child_name):
        """
        get child node.

        args: name of child node.
        """
        return self.children.get(child_name, None)

    def has_child(self, child_name):
        """
        check if child_name exists.

        args: name of a child node.
        """
        return child_name in self.children

    def get_all(self, pattern):
        """
        get all children nodes based on wild card query.

        args: query_pattern.
        """
        matches = []
        patterns = expand_braces(pattern)
        for child in self.children:
            for p in patterns:
                if fnmatch.fnmatch(child, p):
                    matches.append(self.children[child])
        return matches

    def delete(self, child_name):
        """
        provide to pop a child node.

        args: name of child node.
        """
        if self.has_child(child_name):
            self.children.pop(child_name)
            return True
        return False

    def add(self, child_node):
        """
        append a child node.

        args: child node.

        special case:
        Both ZG.seattle and ZG.seattle.zillow exist, we should take care of it as well.
        """
        if child_node.name in self.children:
            if (not self.children[child_node.name].is_leaf) and child_node.is_leaf:
                self.children[child_node.name].is_leaf = child_node.is_leaf
        else:
            self.children[child_node.name] = child_node

    def expand_query(self, query):
        """
        expand a wildcard query.

        args: query.
        return: a list of queries.

        examples: {foo,bar}baz.carbon.cache = [foobaz.carbon.cache, barbaz.carbon.cache]
        """
        sep_index = query.find(self.sep)

        if sep_index < 0:
            return [q.name for q in self.get_all(query) if q.is_leaf]
        else:
            queries = []
            cur_part = query[:sep_index]
            cur_matches = self.get_all(cur_part)
            sub_query = query[sep_index + 1:]
            for match in cur_matches:
                sub_queries = match.expand_query(sub_query)
                for sq in sub_queries:
                    queries.append(".".join([match.name, sq]))
            return queries

    def is_leaf(self):
        """
        return true if the node is leaf.
        """
        return self.is_leaf

    def count(self):
        """
        children count.
        """
        return len(self.children)

    def is_exist(self, query):
        """
        return true if query exist. query match starts from child level/next level.

        query here should not include wildcards.
        """
        if query == "":
            return self.is_leaf

        sep_index = query.find(self.sep)

        if sep_index < 0:
            return self.has_child(query) and self.get(query).is_exist("")
        part = query[:sep_index]
        return self.has_child(part) and self.get(part).is_exist(query[sep_index + 1:])



import pytest


def test_index_has_metric(carbon_index):
    assert carbon_index.has_metric('ZG.zillow.velocity') is True
    assert carbon_index.has_metric('ZG.zillow.velocity.mondev') is True
    assert carbon_index.has_metric('ZG.zillow.product') is False


def test_index_insert(carbon_index):
    assert carbon_index.has_metric('ZG.hotpads') is False
    assert carbon_index.has_metric('ZG.zillow.velocity.perf') is False
    carbon_index.insert('ZG.hotpads')
    carbon_index.insert('ZG.zillow.velocity.perf')
    assert carbon_index.has_metric('ZG.hotpads') is True
    assert carbon_index.has_metric('ZG.zillow.velocity.perf') is True
    carbon_index.insert('ZG.product')
    assert carbon_index.has_metric('ZG.product') is True
    carbon_index.insert('ZG.hotpads.ops')
    assert carbon_index.has_metric('ZG.hotpads.ops') is True
    assert carbon_index.has_metric('ZG.hotpads') is True


def test_index_expand_query_no_wildcard(carbon_index):
    all_matches = carbon_index.expand_query('ZG.zillow.velocity.mondev')
    assert all_matches == ['ZG.zillow.velocity.mondev']


def test_index_expand_query_not_exist(carbon_index):
    all_matches = carbon_index.expand_query('ZG.zillow.product.mortgage')
    assert all_matches == []


def test_index_expand_query_star(carbon_index):
    all_matches = carbon_index.expand_query('ZG.*.product.rental')
    assert sorted(all_matches) == ['ZG.trulia.product.rental', 'ZG.zillow.product.rental']


def test_index_expand_query_brackets(carbon_index):
    all_matches = carbon_index.expand_query('ZG.zi[a-z]low.product.*')
    assert sorted(all_matches) == ['ZG.zillow.product.listing', 'ZG.zillow.product.rental']


def test_index_expand_query_braces(carbon_index):
    all_matches = carbon_index.expand_query('ZG.{trulia,zillow}.product.*')
    assert sorted(all_matches) == ['ZG.trulia.product.rental', 'ZG.zillow.product.listing', 'ZG.zillow.product.rental']


def test_index_delete_empty(carbon_index):
    assert carbon_index.delete('') is False


def test_index_delete_not_exist(carbon_index):
    assert carbon_index.delete('ZG.SF.ops') is False


def test_index_delete_remove(carbon_index):
    assert carbon_index.has_metric('ZG.zillow.product.rental') is True
    assert carbon_index.delete('ZG.zillow.product.rental')
    assert carbon_index.has_metric('ZG.zillow.product.rental') is False
    assert carbon_index.root.get('ZG').get('zillow').get('product').count() == 1

    assert carbon_index.delete('ZG.zillow.product.listing') is True
    assert carbon_index.root.get('ZG').get('zillow').get('product') is None
    assert carbon_index.root.get('ZG').get('zillow').count() == 1



import fnmatch
import re


def _deduplicate(entries):
    yielded = set()
    for entry in entries:
        if entry not in yielded:
            yielded.add(entry)
            yield entry


def match_entries(entries, pattern):
    # First we check for pattern variants (ie. {foo,bar}baz = foobaz or barbaz)
    matching = []

    for variant in expand_braces(pattern):
        matching.extend(fnmatch.filter(entries, variant))

    return list(_deduplicate(matching))


def expand_braces(orig):
    """
    Brace expanding patch for python3 borrowed from:
    https://bugs.python.org/issue9584
    """
    r = r'.*(\{.+?[^\\]\})'
    p = re.compile(r)

    s = orig[:]
    res = list()

    m = p.search(s)
    if m is not None:
        sub = m.group(1)
        open_brace = s.find(sub)
        close_brace = open_brace + len(sub) - 1
        if sub.find(',') != -1:
            for pat in sub.strip('{}').split(','):
                res.extend(
                  expand_braces(s[:open_brace] + pat + s[close_brace + 1:])
                )
        else:
            res.extend(
              expand_braces(s[:open_brace] + sub.replace('}', '\\}') + s[close_brace + 1:])
            )
    else:
        res.append(s.replace('\\}', '}'))

    return list(set(res))
